import argparse

from genreml.model.processing import audio, config
from genreml.model.utils import string_parsing, file_handling, logger


def parse_args():
    """ Parse arguments passed in via the CLI

    :returns a parsed argument object from argparse
    """
    parser = argparse.ArgumentParser()
    # Accept the operation to perform
    parser.add_argument('operation', choices=['download', 'process'], help='''the operation to perform:
    download - search for and download an audio clip for the given song name and artist  
    process - extract data from an audio clip stored in a given file path
    ''')
    parser.add_argument('-s', '--song_name', help='the name of the song to download')
    parser.add_argument('-a', '--artist_name', help='the name of song\'s artist')
    parser.add_argument('-fp', '--file_path', help='the file path of an audio clip or directory of clips to process')
    parser.add_argument('-ex', '--example', action='store_true', help='whether to run an example of feature extraction')
    parser.add_argument('-dp', '--destination_path', help='where to store the results of feature processing')
    parser.add_argument('-e', '--exclude_features', help='a list of feature names to exclude from processing')
    parser.add_argument(
        '-af', '--audio_format', default=config.AudioConfig.AUDIO_FORMAT, help='the format of the audio to process')
    parser.add_argument(
        '-cmap', '--cmap', default=config.DisplayConfig.CMAP,
        help='matplotlib colormap https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html ("None" = default color)'
    )
    parser.add_argument(
        '-fw', '--figure_width', default=config.DisplayConfig.FIGSIZE_WIDTH, help='the width of the figures created')
    parser.add_argument(
        '-fh', '--figure_height', default=config.DisplayConfig.FIGSIZE_HEIGHT, help='the height of the figures created')
    parser.add_argument(
        '-cf', '--checkpoint_frequency', default=config.AudioConfig.CHECKPOINT_FREQUENCY,
        help='how many tracks to process before saving features'
    )
    return parser.parse_args()


def validate_args(args):
    """ Validates the arguments passed in via the CLI """
    if args.operation == 'download' and not (args.song_name and args.artist_name):
        raise RuntimeError('both the song name and artist must be provided to download a clip')
    elif args.operation == 'process' and not (args.file_path or args.example):
        raise RuntimeError(
            'you must either pass in a path to an audio file to process or a path to a directory with audio files')
    elif args.operation == 'process' and args.example and not args.destination_path:
        raise RuntimeError(
            'if running an example feature extraction you must provide a destination path to save the results in'
        )


def set_config(args):
    """ Sets any config attributes provided through CLI """
    # Set the audio format to use in processing
    config.AudioConfig.AUDIO_FORMAT = args.audio_format
    # Set the checkpointing frequency in number of tracks processed
    config.AudioConfig.CHECKPOINT_FREQUENCY = args.checkpoint_frequency


def run(args):
    """ Run the operation as specified via CLI argument """
    # Download clips
    if args.operation == 'download':
       pass
    # Extract features from clips
    elif args.operation == 'process':
        processor = audio.AudioFiles()
        features_to_exclude = string_parsing.str_to_collection(args.exclude_features, set)
        cmap = args.cmap
        if cmap.lower() == "none":
            cmap = None
        # Define the destination path
        if not args.destination_path and not args.example:
            # Use same path as input file_path if a -dp is not provided
            feature_destination_path = file_handling.get_parent_directory(args.file_path)
        else:
            # Otherwise use the path provided via -dp; if running example, always require a destination path
            feature_destination_path = args.destination_path
        # Run the feature extraction
        if args.example:
            processor.extract_sample_fma_features(
                destination_filepath=feature_destination_path, audio_format=args.audio_format)
        else:
            processor.extract_features(args.file_path, feature_destination_path,
                                       features_to_exclude=features_to_exclude,
                                       cmap=cmap, figure_width=float(args.figure_width),
                                       figure_height=float(args.figure_height),
                                       audio_format=args.audio_format
                                       )


def main():
    # Step 1: parse the arguments passed in via the CLI
    args = parse_args()
    # Step 2: validates that the arguments passed in are correct; throws exception if not the case
    validate_args(args)
    # Step 3: Set the config with any eligible inputs to update configs with
    set_config(args)
    # Step 4: Set up logger
    logger.setup_logger()
    # Step 5: run the operation
    run(args)


if __name__ == '__main__':
    main()



# Name: file_handling.py
# Description: defines utility functions to interact with files and directories via the operating system

import os

from pathlib import Path


def get_directory_path(sub_path, base_path=str(Path.home())):
    """ Retrieve the full path that results from concatenating the base_path with the sub_path
     TODO make this more fool-proof and rename

     :param string sub_path: a path to append to the base_path
     :param string base_path: a base_path defaulted to the current machine's home path
     :returns the full concatenated path
    """
    full_path = "{0}{1}".format(base_path, sub_path)
    return full_path


def directory_exists(directory_path):
    """ Checks whether the given path to a directory exists

    :param string directory_path: a file path to a directory
    :returns either True if the directory exists or False otherwise
    """
    return os.path.isdir(directory_path)


def file_exists(filepath: str):
    """ Checks whether the given path to a file exists

    :param filepath: a path to a file
    :returns either True if the file exists or False otherwise
    """
    return os.path.isfile(filepath)


def create_directory(directory_path):
    """ Creates a directory in the given path if it doesn't already exist and opens permissions for the application to
    run successfully

    :param string directory_path: a file path to a directory
    """
    # Check if the directory exists
    if not directory_exists(directory_path):
        # Create it if it doesn't
        os.makedirs(directory_path)
    # Open up permissions to that directory
    os.chmod(directory_path, 0o777)


def get_parent_directory(filepath):
    """ Gets the directory name for a given path; if it's a file then it returns the parent directory and if it's a
    directory then it just returns the same path

    :param string filepath: either a path to a file or directory
    """
    return os.path.dirname(filepath)


def change_directory_to(filepath):
    """ Changes the working directory to the given file path's parent directory

    :param string filepath: a full path to either a file or directory
    """
    # Get the path to the parent directory
    new_path = get_parent_directory(filepath)
    # Change to that new path
    os.chdir(new_path)


def get_filename(filepath):
    """ Extracts the file name from a complete path assuming it's the last item in the path

    :param string filepath: a full path to a file
    :returns the name of the file in the path
    """
    last_item = filepath.split('/')[-1]
    if '.' not in last_item:
        raise ValueError("{0} does not contain a valid file name".format(filepath))
    return last_item


def get_filetype(filepath):
    """ Extracts the type of file from a complete path assuming it's the last item in the path

    :param string filepath: a full path to a file
    :returns the type of the file stored in the path
    """
    file_name = get_filename(filepath)
    return "." + file_name.split(".")[1]



# Name: config.py
# Description: defines configurations for the various components of audio extraction and processing


class AudioConfig:
    # The format to store audio in
    AUDIO_FORMAT = 'mp3'
    # Prefix to save audio features to
    FEATURE_DESTINATION = '/features/'
    # Checkpoint frequency in number of tracks processed
    CHECKPOINT_FREQUENCY = 10


class DisplayConfig:
    # What cmap to use when saving visual features
    # Refer to https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
    CMAP = "Greys"
    # Defines the size of the figures created by display
    FIGSIZE_WIDTH = 10
    FIGSIZE_HEIGHT = 10


class YoutubeExtractionConfig:
    # URL to make request to for audio results
    SEARCH_URL = 'https://www.youtube.com/results'
    # Query parameters to filter requests to search URL by
    SEARCH_QUERY_PARAMS = ['search_query']
    # URL to make request to for actual audio data
    CONTENT_URL = 'https://www.youtube.com/watch'
    # Query parameters to filter requests to content URL by
    CONTENT_QUERY_PARAMS = ['v']
    # Post-processor key to use
    POST_PROCESSOR_KEY = 'FFmpegExtractAudio'
    # Post-processor audio quality
    POST_PROCESSOR_QUALITY = '192'
    # Preferred audio quality name
    PREFERRED_AUDIO_QUALITY = 'bestaudio/best'


class SongExtractorConfig:
    # A list of supported sources by the package
    SUPPORTED_SOURCES = ['youtube']
    # Where to save audio files to
    DESTINATION_FILEPATH = '/Documents/audio-clips/'


class FeatureExtractorConfig:
    # The Librosa features supported by the CLI
    SUPPORTED_FEATURES = ['chroma_stft', 'rms', 'spec_cent', 'spec_bw', 'spec_rolloff', 'zcr', 'mfcc']
    NUMBER_OF_MFCC_COLS = 20
    # How to aggregate the features
    FEATURE_AGGREGATION = ['mean', 'min', 'max', 'std']
    N_FFT = 2048
    HOP_LENGTH = 1024



# Name: audio.py
# Description: defines functionality to process audio from audio files

import logging
import librosa
import os
import pkg_resources
import pandas as pd
import glob
import json

from genreml.model.processing.audio_features import SpectrogramGenerator, LibrosaFeatureGenerator, WavePlotGenerator
from genreml.model.processing.config import AudioConfig, DisplayConfig
from genreml.model.utils import file_handling


class AudioFile(object):

    def __init__(self, file_path: str, audio_signal, sample_rate):
        """ Instantiates an AudioFile object that collects various attributes related to an audio file and exposes
        methods to extract features from that file
        """
        self.file_path = file_path
        self.file_name = file_handling.get_filename(file_path)
        self.audio_type = file_handling.get_filetype(file_path)
        self.audio_signal = audio_signal
        self.sample_rate = sample_rate

    def _get_figure_filepath(self, destination_filepath: str, figure_type: str) -> str:
        return "{0}{1}_{2}".format(
            destination_filepath, figure_type, self.file_name.replace(self.audio_type, ""))

    def _save_figure(self, generator, figure, destination_filepath: str, figure_type: str) -> str:
        path = self._get_figure_filepath(destination_filepath, figure_type)
        logging.info("saving {0} to {1}".format(figure_type, path))
        figure.savefig(path)
        generator.close_img(figure)
        return path + ".png"

    def to_spectrogram(
            self, destination_filepath: str = None, spec_generator=None,
            spectrogram_type: str = "spectrogram", cmap: str = None,
            figure_width: float = DisplayConfig.FIGSIZE_WIDTH, figure_height: float = DisplayConfig.FIGSIZE_HEIGHT):
        """ Extract spectrogram from the audio data and saves the resulting image to the destination path

        :param string destination_filepath: the file path to save the spectrogram to
        :param model.processing.audio_features.SpectrogramGenerator spec_generator: option. spectrogram generator to use
        :param spectrogram_type: the type of spectrogram to generate
        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the spectrogram width in inches
        :param figure_height: the spectrogram height in inches
        :returns the full path location of the saved melspetrogram image
        """
        logging.info("generating {0} for {1}".format(spectrogram_type, self.file_name))
        if not spec_generator:
            spec_generator = SpectrogramGenerator(
                self.audio_signal, self.sample_rate, spectrogram_type=spectrogram_type)
        spectrogram = spec_generator.generate(cmap=cmap, figure_width=figure_width, figure_height=figure_height)
        if destination_filepath:
            _ = self._save_figure(spec_generator, spectrogram, destination_filepath, spectrogram_type)
        return spectrogram

    def to_melspectrogram(self, destination_filepath: str = None, spec_generator=None, cmap: str = None,
                          figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                          figure_height: float = DisplayConfig.FIGSIZE_HEIGHT):
        """ Extract melspectrogram from the audio data and saves the resulting image to the destination path

        :param string destination_filepath: the file path to save the melspectrogram to
        :param model.processing.audio_features.SpectrogramGenerator spec_generator: option. spectrogram generator to use
        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the melspectrogram width in inches
        :param figure_height: the melspectrogram height in inches
        :returns the full path location of the saved melspetrogram image
        """
        return self.to_spectrogram(
            destination_filepath, spec_generator, spectrogram_type="melspectrogram", cmap=cmap,
            figure_width=figure_width, figure_height=figure_height)

    def to_chromagram(self, destination_filepath: str = None, spec_generator=None, cmap: str = None,
                      figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                      figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
                      ):
        """ Extract chromagram from the audio data and saves the resulting image to the destination path

        :param string destination_filepath: the file path to save the chromagram to
        :param model.processing.audio_features.SpectrogramGenerator spec_generator: option. spectrogram generator to use
        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the chromogram width in inches
        :param figure_height: the chromogram height in inches
        :returns the full path location of the saved chromagram image
        """
        return self.to_spectrogram(
            destination_filepath, spec_generator, spectrogram_type="chromagram", cmap=cmap,
            figure_width=figure_width, figure_height=figure_height
        )

    def to_waveplot(self, destination_filepath: str = None, waveplot_generator=None, cmap: str = None,
                    figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                    figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
                    ):
        logging.info("generating waveplot for {0}".format(self.file_name))
        if not waveplot_generator:
            waveplot_generator = WavePlotGenerator(self.audio_signal, self.sample_rate)
        waveplot = waveplot_generator.generate(cmap=cmap, figure_width=figure_width, figure_height=figure_height)
        if destination_filepath:
            _ = self._save_figure(waveplot_generator, waveplot, destination_filepath, "waveplot")
        return waveplot

    def extract_visual_features(
            self, destination_filepath: str = None, cmap: str = DisplayConfig.CMAP, exclusion_set: set = None,
            figure_width: float = DisplayConfig.FIGSIZE_WIDTH, figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
    ) -> list:
        visual_features_to_method_map = {
            "spectrogram": self.to_spectrogram,
            "melspectrogram": self.to_melspectrogram,
            "chromagram": self.to_chromagram,
            "waveplot": self.to_waveplot
        }
        visual_features = []
        for feature_name, feature_method in visual_features_to_method_map.items():
            if feature_name not in exclusion_set:
                visual_features.append(
                    feature_method(
                        destination_filepath, cmap=cmap, figure_width=figure_width, figure_height=figure_height)
                )
        return visual_features

    def extract_features(self,
                         aggregate_features: bool = True,
                         exclusion_set: set = None,
                         feature_generator: LibrosaFeatureGenerator = None) -> tuple:
        """ Extract librosa features from the audio data

        :param aggregate_features - whether to aggregate the features extracted or not
        :param exclusion_set - an optional set of feature names to exclude from the feature generation
        :param feature_generator - an optional generator to use for generating the features
        :returns a dictionary containing the feature names as keys and the data as values
            and a list of feature names in order of processing
        """
        logging.info("generating librosa features for {0}".format(self.file_name))
        if not feature_generator:
            feature_generator = LibrosaFeatureGenerator(
                self.audio_signal, self.sample_rate, aggregate_features, exclusion_set)
        # Extract features
        features, feature_names = feature_generator.generate()
        # Append the identifiers for the current audio file to the feature object
        features['file_name'] = self.file_name
        features['file_path'] = self.file_path
        logging.info("generated {0} features for {1}".format(len(features), self.file_name))
        return features, feature_names

    def __repr__(self):
        return "Audio file of type {0} loaded from {1}".format(self.audio_type, self.file_path)

    def __str__(self):
        return self.__repr__()


class AudioFiles(dict):

    def __init__(self):
        """ Instantiates a collection of AudioFile objects represented as a dictionary where each key is the audio
        file's location on disk and each value the corresponding AudioFile object """
        super(AudioFiles, self).__init__()
        self.visual_features = []
        self.features = []
        self.feature_names = []
        self.features_saved = []

    def _load_file(self, file_location):
        """ Reads in a individual file from the given file_location on disk and keeps a record of those files that
        couldn't be read in

        :param string file_location: the path to a file to read in
        """
        try:
            logging.info("loading audio file from {0}".format(file_location))
            audio_signal, sample_rate = librosa.load(file_location)
            self[file_location] = AudioFile(file_location, audio_signal, sample_rate)
        except Exception as e:
            logging.warning("failed to load audio file from {0} due to {1}".format(file_location, e))
            logging.warning(e)

    def _run_feature_extraction(
            self, audio_file, file_location, destination_filepath=None,
            features_to_exclude=None, cmap=DisplayConfig.CMAP,
            figure_width: float = DisplayConfig.FIGSIZE_WIDTH, figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
    ):
        """ Extracts features from a given AudioFile object representing a file in the given file_location and
        saves the features to destination_filepath

        :param AudioFile audio_file: an AudioFile object
        :param string file_location: the path to the file from with the AudioFile object was instantiated
        :param string destination_filepath: the filepath to save the extracted features to
        :param set features_to_exclude: a collection of feature names to exclude from the final result
        """
        if not features_to_exclude:
            features_to_exclude = set()
        try:
            if destination_filepath:
                destination_filepath = destination_filepath.replace(" ", "")
            self.visual_features.append(audio_file.extract_visual_features(
                destination_filepath, cmap=cmap, exclusion_set=features_to_exclude,
                figure_width=figure_width, figure_height=figure_height
            ))
            feature_dict, feature_names_list = audio_file.extract_features(exclusion_set=features_to_exclude)
            self.features.append(feature_dict)
            self.feature_names.append(feature_names_list)
        except Exception as e:
            logging.warning("failed to extract features from {0}".format(file_location))
            logging.warning(e)
            self.bad_files_extracted[file_location] = e

    def to_df(self):
        """ Store the list of feature dictionaries in a Pandas dataframe where the keys become the columns

        :returns a pandas data frame representation of the audio features
        """
        df = pd.DataFrame()
        record_count = 0
        for idx, feature_dict in enumerate(self.features):
            record_count += 1
            if idx == 0:
                record_to_insert = {key: [val] for key, val in feature_dict.items()}
                df = pd.DataFrame(record_to_insert)
            else:
                df = df.append(feature_dict, ignore_index=True)
        return df, record_count

    def to_csv(self, destination_filepath: str):
        """ Creates a data frame of the features extracted from the current audio files and saves a CSV representation
        of it to the given filepath

        :param destination_filepath - the path to the directory to save the CSV in
        :returns the data frame of features that was saved as CSV
        """
        df, record_count = self.to_df()
        csv_filepath = "{0}{1}{2}.csv".format(destination_filepath, 'feature_data_', str(os.getpid()))
        if file_handling.file_exists(csv_filepath):
            logging.info(
                "appending feature data frame containing {0} records to {1}".format(record_count, csv_filepath))
            df.to_csv(csv_filepath, float_format='%.{}e'.format(10), header=False, index=None, mode='a')
        else:
            logging.info(
                "writing feature data frame containing {0} records to {1}".format(record_count, csv_filepath))
            df.to_csv(csv_filepath, float_format='%.{}e'.format(10), index=None, mode='w')
        return df, csv_filepath

    def _checkpoint_feature_extraction(self, destination_filepath, clear_features=True):
        logging.info("checkpointing progress to {0}".format(destination_filepath))
        _, _ = self.to_csv(destination_filepath)
        self.features_saved.extend(self.features)
        if clear_features:
            self.features = []

    def extract_features(self,
                         file_locations, destination_filepath=None, features_to_exclude=None,
                         load=True, audio_format=AudioConfig.AUDIO_FORMAT, cmap=DisplayConfig.CMAP,
                         figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                         figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
                         ):
        """ Iterates over all of the files in the file_locations, loads them in to extract audio data, and
        generates features

        :param string file_locations: either the location of a single audio file or directory of files to process
        :param string destination_filepath: the location to save features in
        :param set features_to_exclude: a collection of feature names to exclude from the final result
        :param bool load: whether to load the files before extracting the features
        :param string audio_format: the format of the audio files in directory (wav, mp3, etc.)
        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the visual feature figure width in inches
        :param figure_height: the visual feature figure height in inches
        """
        self.features, self.features_saved = [], []
        if destination_filepath:
            destination_filepath = destination_filepath + AudioConfig.FEATURE_DESTINATION
            file_handling.create_directory(destination_filepath)
        # Only load in and process a single file if the given location a file
        if os.path.isfile(file_locations):
            if load:
                self._load_file(file_locations)
            self._run_feature_extraction(
                self[file_locations], file_locations, destination_filepath, features_to_exclude,
                cmap=cmap, figure_width=figure_width, figure_height=figure_height

            )
        # Load in all applicable files if the given location is a directory
        elif os.path.isdir(file_locations):
            logging.info("extracting features from {0} audio files in directory {1}".format(
                audio_format, file_locations))
            if file_locations[-1] == '/':
                file_locations = file_locations[:-1]
            # Retrieve list of files in directory with the matching audio format
            audio_files_in_dir = glob.glob("{0}/*.{1}".format(file_locations, audio_format))
            # Iterate over each file in the directory, load in if applicable, and extract features
            for idx, file in enumerate(audio_files_in_dir):
                if load:
                    self._load_file(file)
                try:
                    self._run_feature_extraction(
                        self[file], file, destination_filepath, features_to_exclude,
                        cmap=cmap, figure_width=figure_width, figure_height=figure_height
                    )
                    # Checkpoint features every AudioConfig.CHECKPOINT_FREQUENCY tracks
                    if (idx + 1) % AudioConfig.CHECKPOINT_FREQUENCY == 0 and destination_filepath:
                        self._checkpoint_feature_extraction(destination_filepath)
                except Exception as e:
                    logging.critical("could not run feature extraction for {0}".format(file))
                    logging.critical(e)
                    if destination_filepath:
                        self._checkpoint_feature_extraction(destination_filepath)
        else:
            raise RuntimeError("file location {0} given to load audio clips from is invalid".format(file_locations))
        if destination_filepath:
            self._checkpoint_feature_extraction(destination_filepath)

    def to_json(self, filepath):
        """ Serializes self as a json in a given file path

        :param string filepath: the file path to save the JSON to
        """
        with open(filepath, 'w') as outfile:
            json.dump(self, outfile)

    def extract_sample_fma_features(self, destination_filepath=None, audio_format=AudioConfig.AUDIO_FORMAT):
        """ Retrieves audio features from sample FMA audio files packaged with the application in genreml/fma_data

        :param str destination_filepath: the optional path to save features to as part of regular feature extraction
        :param string audio_format: the format of the audio files in directory (wav, mp3, etc.)
        """
        path = pkg_resources.resource_filename('genreml', 'fma_data/')
        self.extract_features(
            path, destination_filepath=destination_filepath, audio_format=audio_format)



import glob
import os
import pkg_resources

from genreml.model.processing.audio import AudioFiles


def test_extract_features():
    """ Tests genreml.model.processing.audio.AudioFiles.extract_features method """
    audio_files_processor = AudioFiles()
    sample_data_path = pkg_resources.resource_filename('genreml', 'fma_data/')
    audio_files_processor.extract_sample_fma_features(destination_filepath=os.getcwd())
    files = os.listdir(sample_data_path)
    file_count = len(files)
    assert len(audio_files_processor.features_saved) == file_count
    expected_files = [
        "./features/waveplot*", "./features/melspectrogram*", "./features/chromagram*", "./features/spectrogram*"
    ]
    # There should be as many of the file types above as input file count
    for expected_file in expected_files:
        assert len(glob.glob(expected_file)) == file_count


def test_to_df():
    """ Tests genreml.model.processing.audio.AudioFiles.to_df method """
    audio_files_processor = AudioFiles()
    features = [{"feature_a": 1, "feature_b": 2}, {"feature_a": 2, "feature_b": 3}]
    audio_files_processor.features = features
    df, record_count = audio_files_processor.to_df()
    assert df.shape[0] == record_count \
        and record_count == len(audio_files_processor.features)
    assert len(df.columns) == len(audio_files_processor.features[0])


def test_to_csv():
    """ Tests genreml.model.processing.audio.AudioFiles.to_csv method """
    audio_files_processor = AudioFiles()
    features = {"feature_a": 1, "feature_b": 2}, {"feature_a": 2, "feature_b": 3}
    audio_files_processor.features = features
    df, filepath = audio_files_processor.to_csv("")
    assert df.shape[0] == len(audio_files_processor.features)
    assert os.path.isfile(filepath)
    os.remove(filepath)


def test_feature_checkpointing():
    """ Tests genreml.model.processing.audio.AudioFiles._checkpoint_feature_extraction method """
    audio_files_processor = AudioFiles()
    features = {"feature_a": 1, "feature_b": 2}, {"feature_a": 2, "feature_b": 3}
    audio_files_processor.features = features
    previous_feature_length = len(features)
    audio_files_processor._checkpoint_feature_extraction("")
    assert len(audio_files_processor.features_saved) == previous_feature_length
    assert len(audio_files_processor.features) == 0



# Name: string_parsing.py
# Description: defines utility functions to parse string objects


def str_to_collection(str_to_parse: str, collection_to_parse_to: any, delim: str = ','):
    """ Converts a given string with the given delimeter into an object of the given type

    :param str_to_parse - the string to parse into an object
    :param collection_to_parse_to - the class of the collection to parse the string to
    :param delim - the delimeter that separates values in the string
    :returns a collection of values parsed from the given string
    """
    if collection_to_parse_to not in (set, dict, list):
        raise RuntimeError(
            "a string cannot be parsed into the given collection of {0}", format(str(collection_to_parse_to)))
    if not str_to_parse:
        str_list = []
    else:
        str_list = str_to_parse.split(delim)
    return collection_to_parse_to(str_list)



# Name: audio_features.py
# Description: defines how to extract different types of features from audio files

import librosa
import librosa.display
import logging
import matplotlib.pyplot as plt
import numpy as np
import warnings

from abc import ABC, abstractmethod

from genreml.model.processing.config import FeatureExtractorConfig, DisplayConfig
from genreml.model.processing.display import VisualDataMixin


class FeatureGenerator(ABC):
    """ Abstract feature generation class defines common functionality and interface for all feature generators """

    def __init__(self, audio_signal: np.array, sample_rate: np.array, features_to_exclude: any, config):
        self.audio_signal = audio_signal
        self.sample_rate = sample_rate
        self.features_to_exclude = features_to_exclude
        if not features_to_exclude:
            self.features_to_exclude = set()
        self.config = config

    @abstractmethod
    def generate(self):
        pass


class SpectrogramGenerator(FeatureGenerator, VisualDataMixin):
    """ Generates a spectrogram image from audio data

    https://librosa.org/doc/0.7.2/generated/librosa.feature.melspectrogram.html
    https://medium.com/analytics-vidhya/understanding-the-mel-spectrogram-fca2afa2ce53
    """

    def __init__(self,
                 audio_signal: np.ndarray, sample_rate: np.ndarray,
                 spectrogram_type: str = "melspectrogram", config=FeatureExtractorConfig):
        super().__init__(audio_signal, sample_rate, features_to_exclude=None, config=config)
        self.spectrogram_type = spectrogram_type

    def _create_db_melspectrogram_data(self, audio_signal: np.ndarray, sample_rate: np.ndarray) -> np.ndarray:
        """ Creates the data for a decibel melspectrogram using librosa

        :param audio_signal: an audio time-series
        :param sample_rate: the sampling rate of the audio time-series
        :returns a numpy array containing the data to plot as a melspectrogram
        """
        mel_spect = librosa.feature.melspectrogram(
            y=audio_signal, sr=sample_rate, n_fft=self.config.N_FFT, hop_length=self.config.HOP_LENGTH
        )
        return librosa.power_to_db(mel_spect)

    def _create_db_spectrogram_data(self, audio_signal: np.ndarray) -> np.ndarray:
        """ Creates the data for a decibel spectrogram using librosa

        :param audio_signal: an audio time-series
        :returns a numpy array containing the data to plot as a spectrogram
        """
        spectrogram_data = np.abs(
            librosa.stft(y=audio_signal, n_fft=self.config.N_FFT, hop_length=self.config.HOP_LENGTH))
        return librosa.amplitude_to_db(spectrogram_data, ref=np.max)

    def _create_chromagram_data(self, audio_signal: np.ndarray, sample_rate: np.ndarray) -> np.ndarray:
        """ Generates the data from a chromagram using librosa

        :param audio_signal: an audio time-series
        :param sample_rate: the sampling rate of the audio time-series
        :returns a numpy array containing the data to plot as a chromagram
        """
        chromagram = librosa.feature.chroma_stft(audio_signal, sample_rate, hop_length=self.config.HOP_LENGTH)
        return chromagram

    def generate(self, cmap: str = None,
                 figure_width: float = DisplayConfig.FIGSIZE_WIDTH, figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
                 ) -> plt.figure:
        """ Generates a spectrogram by calling the different component methods of this object

        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the spectrogram width in inches
        :param figure_height: the spectrogram height in inches
        :returns a matplotlib.pyplot.figure object visualizing the spectrogram data
        """
        if self.spectrogram_type == "melspectrogram":
            mel_spect = self._create_db_melspectrogram_data(self.audio_signal, self.sample_rate)
            norm_mel_spect = self.normalize(mel_spect)
            eight_bit_spectrogram = self.convert_pixels_to_8_bits(norm_mel_spect)
            transformed_spectrogram = self.flip_and_invert(eight_bit_spectrogram)
        elif self.spectrogram_type == "chromagram":
            chromagram = self._create_chromagram_data(self.audio_signal, self.sample_rate)
            transformed_spectrogram = chromagram
        else:
            spect = self._create_db_spectrogram_data(self.audio_signal)
            transformed_spectrogram = spect
        return self.display_data(
            transformed_spectrogram, cmap=cmap, figure_width=figure_width, figure_height=figure_height)


class WavePlotGenerator(FeatureGenerator, VisualDataMixin):
    """ Generates a waveplot image from audio data

    https://librosa.org/doc/0.7.2/generated/librosa.feature.melspectrogram.html
    https://medium.com/analytics-vidhya/understanding-the-mel-spectrogram-fca2afa2ce53
    """

    def __init__(self, audio_signal: np.ndarray, sample_rate: np.ndarray, config=FeatureExtractorConfig):
        super().__init__(audio_signal, sample_rate, features_to_exclude=None, config=config)

    def generate(self, cmap: str = None,
                 figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                 figure_height: float = DisplayConfig.FIGSIZE_HEIGHT) -> plt.figure:
        """ Generates a wave plot using librosa

        :param cmap: https://matplotlib.org/3.3.2/api/_as_gen/matplotlib.axes.Axes.imshow.html
        :param figure_width: the waveplot width in inches
        :param figure_height: the waveplot height in inches

        :returns a matplotlib.pyplot.figure object visualizing the wave plot data
        """
        # TODO: this is not adding axes for some reason
        fig, ax = self.create_display_figure(
            frameon=True, display_axes=True, x_axis_name="Time",
            figure_width=figure_width, figure_height=figure_height)
        librosa.display.waveplot(self.audio_signal, self.sample_rate)
        return fig


class LibrosaFeatureGenerator(FeatureGenerator):

    def __init__(self,
                 audio_signal: np.array, sample_rate: np.array, aggregate_features: bool = True,
                 features_to_exclude: list = None, config=FeatureExtractorConfig):
        """ Constructor capturing the raw audio_signal and sample_rate data extracted from librosa audio processing
        to generate features from that data
        """
        super().__init__(audio_signal, sample_rate, features_to_exclude, config)
        # Ignore Librosa warning regarding PySoundFile
        warnings.filterwarnings('ignore', module='librosa')
        self.aggregate_features = aggregate_features
        self.supported_features = self.config.SUPPORTED_FEATURES
        self.aggregations = self.config.FEATURE_AGGREGATION

    def _get_feature_function(self, feature: str):
        """ Factory method for librosa function corresponding to a given feature name

        :param feature - the name of the feature for which to retrieve the librosa function
        :returns the librosa function object and inputs it requires to generate the given feature name
        """
        # Some librosa functions require both audio signal and sample rate, some just audio signal, etc.
        input_types = [
            {'y': self.audio_signal, 'sr': self.sample_rate},
            {'y': self.audio_signal},
            {'y': self.audio_signal, 'sr': self.sample_rate, 'n_mfcc': self.config.NUMBER_OF_MFCC_COLS},
        ]
        features = {
            'chroma_stft': (librosa.feature.chroma_stft, input_types[0]),
            'rms': (librosa.feature.rms, input_types[1]),
            'spec_cent': (librosa.feature.spectral_centroid, input_types[0]),
            'spec_bw': (librosa.feature.spectral_bandwidth, input_types[0]),
            'spec_rolloff': (librosa.feature.spectral_rolloff, input_types[0]),
            'zcr': (librosa.feature.zero_crossing_rate, input_types[1]),
            'mfcc': (librosa.feature.mfcc, input_types[2])
        }
        try:
            return features[feature]
        except KeyError:
            logging.critical("feature {0} is not associated with a librosa feature function".format(feature))

    def _extract_features(self) -> tuple:
        """ Extracts the supported features from Librosa and returns the data

        :returns a dictionary containing the names of the features extracted as keys and the data arrays as values
            and a list of feature names in order of processing
        """
        # Define a feature dictionary in which to store feature aggregations
        feature_dict = {}
        feature_names = []
        for feature in self.supported_features:
            # Skip feature generation for any features that need to be excluded
            if feature in self.features_to_exclude:
                continue
            try:
                # Get the function and inputs to use for the given feature
                librosa_function, inputs = self._get_feature_function(feature)
                # Extract the feature data
                feature_data = librosa_function(**inputs)
                feature_dict[feature] = feature_data
                feature_names.append(feature)
            except Exception as e:
                logging.critical("could not extract feature {0}".format(feature))
                raise
        return feature_dict, feature_names

    def _aggregate_features(self, feature_dict: dict) -> tuple:
        """ Aggregates the features in given dictionary of features according to supported aggregations

        :param feature_dict - a dictionary containing feature names to data arrays
        :returns a new dictionary containing the names of the feature aggregations as keys and agg results as values
            and a list of feature names in order of processing
        """
        aggregated_features = {}
        aggregation_functions = {'mean': np.mean, 'min': np.min, 'max': np.max, 'std': np.std}
        feature_names = []
        for feature, data in feature_dict.items():
            # Aggregate the feature data and store the aggregations in the aggregated features dictionary
            if feature != 'mfcc':
                for aggregation in self.aggregations:
                    if aggregation not in aggregation_functions:
                        raise ValueError(
                            "aggregation {0} is not associated with a valid aggregation function".format(aggregation))
                    # Apply the aggregation result and store it
                    aggregation_function = aggregation_functions[aggregation]
                    feature_name = '{0}-{1}'.format(feature, aggregation)
                    aggregated_features[feature_name] = aggregation_function(data)
                    feature_names.append(feature_name)
            else:
                # Other features can't be aggregated
                for mfcc_col in range(self.config.NUMBER_OF_MFCC_COLS):
                    feature_name = 'mfcc{0}'.format(mfcc_col)
                    aggregated_features[feature_name] = aggregation_functions['mean'](data[mfcc_col])
                    feature_names.append(feature_name)
        return aggregated_features, feature_names

    def generate(self) -> tuple:
        """ Runs the main feature generator logic and returns a dictionary with the processed features

        :returns a dictionary with feature names or aggregations as keys and data/aggregation results as values
            and a list of feature names in order of processing
        """
        # Extract all of the features in config-based supported features
        features, feature_names = self._extract_features()
        # If required, aggregate the features extracted
        if self.aggregate_features:
            features, feature_names = self._aggregate_features(features)
        return features, feature_names



import pytest

from genreml.model import __main__


class MockArgs(object):
    """ Args object mimicking ArgParse """
    operation = None
    artist_name = None
    song_name = None
    file_path = None
    example = False
    audio_format = None
    checkpoint_frequency = None
    destination_path = None


def test_validate_args():
    """ Tests genreml.model.__main__.validate_args function """
    # Test nominal download operation
    mock_args = MockArgs()
    mock_args.operation = 'download'
    mock_args.song_name = 'some_song'
    mock_args.artist_name = 'some_artist'
    __main__.validate_args(mock_args)
    # Missing required download attributes
    mock_args.song_name = None
    with pytest.raises(RuntimeError):
        __main__.validate_args(mock_args)
    mock_args.song_name = 'some_song'
    mock_args.artist_name = None
    with pytest.raises(RuntimeError):
        __main__.validate_args(mock_args)
    # Test nominal process operation
    mock_args.operation = 'process'
    mock_args.example = True
    mock_args.destination_path = 'some_path'
    __main__.validate_args(mock_args)
    mock_args.example = False
    mock_args.file_path = 'some_file_path'
    mock_args.destination_path = 'some_path'
    __main__.validate_args(mock_args)
    mock_args.destination_path = None
    __main__.validate_args(mock_args)
    # Missing required process attributes
    mock_args.example = True
    with pytest.raises(RuntimeError):
        __main__.validate_args(mock_args)
    mock_args.example = False
    mock_args.file_path = False
    with pytest.raises(RuntimeError):
        __main__.validate_args(mock_args)



from genreml.model.processing import config


def test_audio_config():
    """ Tests that genreml.model.processing.config.AudioConfig has all the required attributes """
    checkpoint_frequency = config.AudioConfig.CHECKPOINT_FREQUENCY
    audio_format = config.AudioConfig.AUDIO_FORMAT
    feature_destination = config.AudioConfig.FEATURE_DESTINATION
    assert checkpoint_frequency and audio_format and feature_destination


def test_youtube_extractor_config():
    """ Tests that genreml.model.processing.config.YoutubeExtractorConfig has all the required attributes """
    post_processor_quality = config.YoutubeExtractionConfig.POST_PROCESSOR_QUALITY
    post_processor_key = config.YoutubeExtractionConfig.POST_PROCESSOR_KEY
    preferred_audio_quality = config.YoutubeExtractionConfig.PREFERRED_AUDIO_QUALITY
    content_query_params = config.YoutubeExtractionConfig.CONTENT_QUERY_PARAMS
    content_url = config.YoutubeExtractionConfig.CONTENT_URL
    search_query_params = config.YoutubeExtractionConfig.SEARCH_QUERY_PARAMS
    search_url = config.YoutubeExtractionConfig.SEARCH_URL
    assert search_url and search_query_params and content_url and content_query_params and preferred_audio_quality \
        and post_processor_key and post_processor_quality and preferred_audio_quality


def test_song_extractor_config():
    """ Tests that genreml.model.processing.config.SongExtractorConfig has all the required attributes """
    supported_sources = config.SongExtractorConfig.SUPPORTED_SOURCES
    destination_filepath = config.SongExtractorConfig.DESTINATION_FILEPATH
    assert supported_sources and destination_filepath


def test_feature_extractor_config():
    """ Tests that genreml.model.processing.config.FeatureExtractorConfig has all the required attributes """
    mfcc_cols = config.FeatureExtractorConfig.NUMBER_OF_MFCC_COLS
    feature_aggregation = config.FeatureExtractorConfig.FEATURE_AGGREGATION
    supported_features = config.FeatureExtractorConfig.SUPPORTED_FEATURES
    assert mfcc_cols and feature_aggregation and supported_features



# Name: display.py
# Description: defines common display functionality for matplot graphics

import matplotlib.pyplot as plt
import numpy as np

from genreml.model.processing.config import DisplayConfig


class VisualDataMixin(object):
    """ Defines common functionality for visualization of features """

    @staticmethod
    def normalize(visual_data: np.array) -> np.array:
        return 255 * ((visual_data - visual_data.min()) /
                      (visual_data.max() - visual_data.min()))

    @staticmethod
    def convert_pixels_to_8_bits(visual_data: np.array) -> np.array:
        return visual_data.astype(np.uint8)

    @staticmethod
    def flip_and_invert(visual_data: np.array) -> np.array:
        flipped_img = np.flip(visual_data, axis=0)
        return 255 - flipped_img

    @staticmethod
    def create_display_figure(frameon: bool = False, display_axes: bool = False,
                              x_axis_name: str = None, y_axis_name: str = None,
                              figure_width: float = DisplayConfig.FIGSIZE_WIDTH,
                              figure_height: float = DisplayConfig.FIGSIZE_HEIGHT) -> tuple:
        print("Creating display figure with {0},{1}".format(figure_width, figure_height))
        fig = plt.figure(frameon=frameon, figsize=(figure_width, figure_height))
        ax = plt.Axes(fig, [0., 0., 1., 1.])
        if not display_axes:
            ax.set_axis_off()
        else:
            ax.set(xlabel=x_axis_name, ylabel=y_axis_name)
        fig.add_axes(ax)
        return fig, ax

    @staticmethod
    def display_data(
            visual_data: np.array, frameon: bool = False, cmap: str = DisplayConfig.CMAP,
            display_axes: bool = False, x_axis_name: str = None, y_axis_name: str = None,
            figure_width: float = DisplayConfig.FIGSIZE_WIDTH, figure_height: float = DisplayConfig.FIGSIZE_HEIGHT
    ) -> plt.figure:
        """ Displays the given data in a matplotlib figure and returns the figure object """
        fig, ax = VisualDataMixin.create_display_figure(
            frameon, display_axes, x_axis_name, y_axis_name, figure_width=figure_width, figure_height=figure_height)
        ax.imshow(visual_data, aspect='auto', cmap=cmap)
        return fig

    @staticmethod
    def close_img(fig: plt.figure) -> None:
        """ Closes a matplotlib figure """
        plt.close(fig)



import pytest

from genreml.model.utils import string_parsing


def test_str_to_collection():
    """ Tests genreml.model.utils.string_parsing.str_to_collection function """
    sample_str = 'something,in,a,string'
    # Test nominal operation
    assert string_parsing.str_to_collection(sample_str, set) == {'something', 'in', 'a', 'string'}
    assert string_parsing.str_to_collection(sample_str, list) == ['something', 'in', 'a', 'string']
    # Test wrong delim type
    assert string_parsing.str_to_collection(sample_str, list, delim='\t') == [sample_str]
    # Test unsupported collection type
    with pytest.raises(RuntimeError):
        string_parsing.str_to_collection(sample_str, str)



import aiohttp
import asyncio
import hashlib
import json
import os
import uuid
from quart import Quart, request, jsonify, abort, redirect, url_for, url_for, render_template


app = Quart(__name__, static_folder='./static', static_url_path='/static')
event_loop = asyncio.get_event_loop()
filestore_directory = "/filestore"


@app.route('/uploadmultiple', methods=['POST'])
async def file_upload_handler():
    global filestore_directory
    files = (await request.files).getlist("multiFileUploadForm")
    for f in files:
        data = f.read()
        if type(data) == bytes:
            file_hash = hashlib.sha256(data).hexdigest()
            # currently the files are simply written to disk.
            # log term data hash + prediction stored in a database
            # or temporary in memory object store may be a good idea
            # so model is only run on new unknown data.
            # if performance can be guaranteed to be good enough
            # this will be replaced with API call to prediction
            # engine. Otherwise, a token will have to be returned
            # for polling until confirmation that audio processing
            # and prediction work is complete.
            write_path = "/".join([filestore_directory, file_hash])
            if not os.path.isfile(write_path):
                with open(write_path, "wb") as fh:
                    fh.write(data)
    return redirect(url_for('index'))


@app.route('/')
async def index():
    """ Server route for the app's landing page """
    return await render_template('index.html')


import os
import pytest

from genreml.model.utils import file_handling


def test_get_directory_path():
    """ Tests genreml.model.utils.file_handling.get_directory_path function """
    # Nominal operation
    base_path = '/Users/JaneDoe/Documents'
    sub_path = '/audio-clips/'
    assert file_handling.get_directory_path(sub_path, base_path) == '/Users/JaneDoe/Documents/audio-clips/'


def test_directory_exists():
    """ Tests genreml.model.utils.file_handling.directory_exists function """
    # Nominal operation
    pid = os.getpid()
    directory_name = './test_dir_exists' + str(pid)
    os.mkdir(directory_name)
    assert file_handling.directory_exists(directory_name)
    os.rmdir(directory_name)
    assert not file_handling.directory_exists(directory_name)


def test_file_exists():
    """ Tests genreml.model.utils.file_handling.file_exists function """
    # Nominal operation
    pid = os.getpid()
    file_name = './test_file' + str(pid)
    with open(file_name, 'w') as f:
        f.write("test file content")
    assert file_handling.file_exists(file_name)
    os.remove(file_name)
    assert not file_handling.file_exists(file_name)


def test_create_directory():
    """ Tests genreml.model.utils.file_handling.create_directory function """
    # Nominal operation
    pid = os.getpid()
    directory_name = './test_dir_create' + str(pid)
    file_handling.create_directory(directory_name)
    assert file_handling.directory_exists(directory_name)
    # Trying to create directory again should do nothing
    file_handling.create_directory(directory_name)
    os.rmdir(directory_name)
    assert not file_handling.directory_exists(directory_name)
    file_handling.create_directory(directory_name)
    assert file_handling.directory_exists(directory_name)
    os.rmdir(directory_name)


def test_get_parent_directory():
    """ Tests genreml.model.utils.file_handling.get_parent_directory function """
    # Nominal operation
    pid = os.getpid()
    base_path = './test_dir_parent'
    sub_path = '/audio-clips{0}/'.format(str(pid))
    file_name = 'some_file.txt'
    full_path = file_handling.get_directory_path(sub_path + file_name, base_path)
    parent_dir_path = file_handling.get_directory_path(sub_path, base_path)
    expected_result = sub_path.split('/')[1]
    file_handling.create_directory(full_path)
    assert file_handling.get_parent_directory(full_path).split('/')[-1] == expected_result
    assert file_handling.get_parent_directory(parent_dir_path).split('/')[-1] == expected_result


def test_get_filename():
    """ Tests genreml.model.utils.file_handling.get_filename function """
    # Nominal operation
    pid = os.getpid()
    base_path = './test_dir_parent'
    sub_path = '/audio-clips{0}/'.format(str(pid))
    file_name = 'some_file.txt'
    full_path = file_handling.get_directory_path(sub_path + file_name, base_path)
    parent_dir_path = file_handling.get_directory_path(sub_path, base_path)
    assert file_handling.get_filename(full_path) == file_name
    # Bad input
    with pytest.raises(ValueError):
        file_handling.get_filename(parent_dir_path)


def test_get_filetype():
    """ Tests genreml.model.utils.file_handling.get_filetype function """
    # Nominal operation
    some_path_1 = "some_path_1/some_file.txt"
    some_path_2 = "some_path_2/some_file.mp3"
    some_path_3 = "some_path_3/some_sub_path_3/"
    assert file_handling.get_filetype(some_path_1) == '.txt'
    assert file_handling.get_filetype(some_path_2) == '.mp3'
    # Bad input
    with pytest.raises(ValueError):
        file_handling.get_filetype(some_path_3)
